# -*- coding: utf-8 -*-
"""Base Tensor class."""
from typing import ClassVar
from typing import Dict
from typing import Optional

from mrsimulator.utils.parseable import Parseable
from pydantic import Field

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"


class SymmetricTensor(Parseable):
    r"""Base SymmetricTensor class representing the traceless symmetric part of an
    irreducible second-rank tensor.

    Attributes
    ----------

    zeta: float (optional).
        The anisotropy parameter of the nuclear shielding tensor, in ppm, expressed
        using the Haeberlen convention. The default value is None.

        Example
        -------

        >>> shielding = SymmetricTensor()
        >>> shielding.zeta = 10

    Cq: float (optional).
        The quadrupolar coupling constant, in Hz, derived from the electric field
        gradient tensor. The default value is None.

        Example
        -------

        >>> efg = SymmetricTensor()
        >>> efg.Cq = 10e6

    eta: float (optional).
        The asymmetry parameter of the SymmetricTensor expressed using the Haeberlen
        convention. The default value is None.

        Example
        -------

        >>> shielding.eta = 0.1
        >>> efg.eta = 0.5

    alpha: float (optional).
        Euler angle, :math:`\alpha`, in radians. The default value is None.

        Example
        -------

        >>> shielding.alpha = 0.15
        >>> efg.alpha = 1.5

    beta: float (optional).
        Euler angle, :math:`\beta`, in radians. The default value is None.

        Example
        -------

        >>> shielding.beta = 3.1415
        >>> efg.beta = 1.1451

    gamma: float (optional).
        Euler angle, :math:`\gamma`, in radians. The default value is None.

        Example
        -------

        >>> shielding.gamma = 2.1
        >>> efg.gamma = 0

    Example
    -------

    >>> shielding = SymmetricTensor(zeta=10, eta=0.1, alpha=0.15, beta=3.14, gamma=2.1)
    >>> efg = SymmetricTensor(Cq=10e6, eta=0.5, alpha=1.5, beta=1.1451, gamma=0)
    """

    zeta: float = None
    Cq: float = None
    D: float = None
    eta: float = Field(default=None, ge=0.0, le=1.0)
    alpha: float = None
    beta: float = None
    gamma: float = None

    property_unit_types: ClassVar = {
        "zeta": ["dimensionless", "frequency"],
        "Cq": "frequency",
        "D": "frequency",
        "alpha": "angle",
        "beta": "angle",
        "gamma": "angle",
    }
    property_default_units: ClassVar = {
        "zeta": ["ppm", "Hz"],
        "Cq": "Hz",
        "D": "Hz",
        "alpha": "rad",
        "beta": "rad",
        "gamma": "rad",
    }
    property_units: Dict = {
        "zeta": "ppm",
        "Cq": "Hz",
        "D": "Hz",
        "alpha": "rad",
        "beta": "rad",
        "gamma": "rad",
    }

    # Deprecated
    # def to_freq_dict(self, larmor_frequency: float) -> dict:
    #     """
    #     Serialize the SymmetricTensor object to a JSON compliant python dictionary
    #     where the attribute values are numbers expressed in default units. The default
    #     unit for attributes with respective dimensionalities are:
    #     - frequency: `Hz`
    #     - angle: `rad`

    #     Args:
    #         float larmor_frequency: The larmor frequency in MHz.

    #     Return:
    #         A python dict
    #     """
    #     temp_dict = self.dict()
    #     if temp_dict["zeta"] is not None:
    #         temp_dict["zeta"] *= larmor_frequency
    #     temp_dict.pop("property_units")
    #     return temp_dict


class AntisymmetricTensor(Parseable):
    """Base SymmetricTensor class representing the traceless symmetric part of an
    irreducible second-rank tensor.

    Attributes:
        zeta: The anisotropy parameter of the AntiSymmetricTensor expressed using
            the Haeberlen convention. The default value is None.
        alpha: Euler angle, alpha, given in radian. The default value is None.
        beta: Euler angle, beta, given in radian. The default value is None.
    """

    zeta: Optional[float]
    alpha: Optional[float]
    beta: Optional[float]

    property_unit_types: ClassVar = {
        "zeta": "dimensionless",
        "alpha": "angle",
        "beta": "angle",
    }
    property_default_units: ClassVar = {"zeta": "ppm", "alpha": "rad", "beta": "rad"}
    property_units: Dict = {"zeta": "ppm", "alpha": "rad", "beta": "rad"}

    # Deprecated
    # def to_freq_dict(self, larmor_frequency: float) -> dict:
    #     """
    #     Serialize the AntisymmetricTensor object to a JSON compliant python dictionary
    #     where the attribute values are numbers expressed in default units. The default
    #     unit for attributes with respective dimensionalities are:
    #     - frequency: `Hz`
    #     - angle: `rad`

    #     Args:
    #         float larmor_frequency: The larmor frequency in MHz.

    #     Return:
    #         Python dict
    #     """
    #     temp_dict = self.dict()
    #     if temp_dict["zeta"] is not None:
    #         temp_dict["zeta"] *= larmor_frequency
    #     temp_dict.pop("property_units")
    #     return temp_dict



# -*- coding: utf-8 -*-
import csdmpy as cp
import numpy as np
import pytest
from mrsimulator import signal_processing as sp

__author__ = "Maxwell C. Venetos"
__email__ = "maxvenetos@gmail.com"


def setup_read_write(operation, py_dict, fn):
    # class to dict with units
    dict_ = operation.json()
    assert dict_ == py_dict

    # read from dictionary
    b = fn.parse_dict_with_units(dict_)
    assert operation == b

    processor = sp.SignalProcessor.parse_dict_with_units({"operations": [dict_]})
    assert operation == processor.operations[0]


def test_01():
    post_sim = sp.SignalProcessor()
    operations = [
        sp.IFFT(),
        sp.apodization.Gaussian(FWHM="12 K", dim_index=0, dv_index=0),
        sp.FFT(),
    ]

    post_sim.operations = operations

    with pytest.raises(ValueError, match="The data must be a CSDM object."):
        post_sim.apply_operations([])

    data = cp.as_csdm(np.arange(20))
    data.x[0] = cp.LinearDimension(count=20, increment="10 K")
    post_sim.apply_operations(data)

    # to dict with units
    dict_ = post_sim.json()
    assert dict_ == {
        "operations": [
            {"dim_index": 0, "function": "IFFT"},
            {
                "function": "apodization",
                "type": "Gaussian",
                "FWHM": "12.0 K",
                "dim_index": 0,
                "dv_index": 0,
            },
            {"dim_index": 0, "function": "FFT"},
        ],
    }

    # parse dict with units
    post_sim_2 = sp.SignalProcessor.parse_dict_with_units(dict_)

    assert post_sim.operations == post_sim_2.operations


def generate_data():
    dv1 = cp.as_dependent_variable(np.random.rand(20))
    dv2 = cp.as_dependent_variable(np.random.rand(20))
    dv3 = cp.as_dependent_variable(np.random.rand(20))
    dim = cp.as_dimension(np.arange(20))
    return cp.CSDM(dependent_variables=[dv1, dv2, dv3], dimensions=[dim])


def test_scale():
    data_in = generate_data()
    PS_0 = [sp.Scale(factor=10)]
    operator = sp.SignalProcessor(operations=PS_0)
    data_out = operator.apply_operations(data=data_in.copy())
    _, y0, y1, y2 = data_in.to_list()
    _, y0_, y1_, y2_ = data_out.to_list()

    for in_, out_ in zip([y0, y1, y2], [y0_, y1_, y2_]):
        assert np.allclose(out_.max() / in_.max(), 10), "Scaling failed"


def test_linear():
    data_in = generate_data()
    PS_0 = [sp.Linear(amplitude=4.1, offset=10)]
    operator = sp.SignalProcessor(operations=PS_0)
    data_out = operator.apply_operations(data=data_in.copy())
    _, y0, y1, y2 = data_in.to_list()
    _, y0_, y1_, y2_ = data_out.to_list()

    for in_, out_ in zip([y0, y1, y2], [y0_, y1_, y2_]):
        assert np.allclose((out_.max() - 10) / in_.max(), 4.1), "Offset failed"



# -*- coding: utf-8 -*-
from typing import Dict
from typing import List
from typing import Optional

from mrsimulator.transition import Transition
from pydantic import BaseModel
from pydantic import Field

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class TransitionQuery(BaseModel):
    """Base TransitionQuery class.

    Attributes
    ----------

    P:
        A dict of channels, with each channel as a list of p symmetry functions per
        site. Here p = Δm is the difference between spin quantum numbers of the final
        and initial states.

        Example
        -------

        >>> method = Method2D()
        >>> method.spectral_dimensions[0].events[0].transition_query.P = {
        ...     'channel-1': [[-1]]
        ... }

    D:
        A dict of channels, with each channel as a list of d symmetry functions per
        site. Here p = Δm is the difference between spin quantum numbers of the final
        and initial states.

        Example
        -------

        >>> method.spectral_dimensions[0].events[0].transition_query.D = {
        ...     'channel-1': [[0]]
        ... }
    """

    P: Optional[Dict] = {"channel-1": [[-1.0]]}
    D: Optional[Dict] = Field(default=None)
    F: Optional[Dict] = Field(default=None)
    transitions: List[Transition] = None

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    def json(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object, where
        the attribute value with physical quantity is expressed as a string with a
        value and a unit."""
        return {k: v for k, v in self.dict().items() if v is not None}



# -*- coding: utf-8 -*-
import numpy as np
from mrsimulator.method.spectral_dimension import SpectralDimension

from .base import BaseNamedMethod2D

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class SSB2D(BaseNamedMethod2D):
    r"""Simulating a sheared 2D finite to infinite speed MAS correlation spectum.

    For spin I=1/2, the infinite speed MAS is the isotropic dimension.

    Args:
        channels: A list of isotope symbols over which the method will be applied.
        spectral_dimensions: A list of python dict. Each dict is contains keywords that
            describe the coordinates along a spectral dimension. The keywords along with
            its definition are:

            - count:
                An optional integer with the number of points, :math:`N`, along the
                dimension. The default value is 1024.
            - spectral_width:
                An `optional` float with the spectral width, :math:`\Delta x`, along the
                dimension in units of Hz. The default is 25 kHz.
            - reference_offset:
                An `optional` float with the reference offset, :math:`x_0` along the
                dimension in units of Hz. The default value is 0 Hz.
            - origin_offset:
                An `optional` float with the origin offset (Larmor frequency) along the
                dimension in units of Hz. The default value is None.

        rotor_frequency: An `optional` float containing the sample spinning frequency
            :math:`\nu_r`, in units of Hz. The default value is ``0``.
        magetic_flux_density: An `optional` float containing the macroscopic magnetic
            flux density, :math:`H_0`, of the applied external magnetic field in units
            of T. The default value is 9.4.
        rotor_angle: An `optional` float containing the angle between the sample
            rotation axis and the applied external magnetic field, :math:`\theta`, in
            units of rad. The default value is 0.9553166, i.e. the magic angle.

    Return:
        A :class:`~mrsimulator.Method` instance.

    Example:
        >>> method = SSB2D(
        ...     channels=["13C"],
        ...     magnetic_flux_density=7,  # in T
        ...     rotor_frequency=1500, # in Hz
        ...     spectral_dimensions=[
        ...         {
        ...             "count": 16,
        ...             "spectral_width": 16*1500,  # in Hz (= count * rotor_frequency)
        ...             "reference_offset": -5e3,  # in Hz
        ...             "label": "Sideband dimension",
        ...         },
        ...         {
        ...             "count": 512,
        ...             "spectral_width": 1e4,  # in Hz
        ...             "reference_offset": -4e3,  # in Hz
        ...             "label": "Isotropic dimension",
        ...         },
        ...     ],
        ... )
        >>> sys = SpinSystem(sites=[Site(isotope='13C')])
        >>> method.get_transition_pathways(sys)
        [|-0.5⟩⟨0.5| ⟶ |-0.5⟩⟨0.5|]
    """

    def __new__(cls, **kwargs):
        # check for rotor frequency
        if "spectral_dimensions" not in kwargs:
            kwargs["spectral_dimensions"] = [{}, {}]
        token = isinstance(kwargs["spectral_dimensions"][0], SpectralDimension)
        error = f"`rotor_frequency` cannot be zero for {__class__.__name__} method."

        if not token:
            check = [
                "rotor_frequency" not in ev
                for item in kwargs["spectral_dimensions"]
                if "events" in item
                for ev in item["events"]
            ]
            if "rotor_frequency" not in kwargs and np.all(check):
                raise ValueError(error)
            if kwargs["rotor_frequency"] == 0:
                raise ValueError(error)

            rotor_frequency = kwargs.pop("rotor_frequency")

        method = super().__new__(cls, **kwargs)

        if not token:
            # update rotor frequency for the first dimension to user defined speed.
            method.spectral_dimensions[0].events[0].rotor_frequency = rotor_frequency

        # check that the spectral width for the first dimension is equal to the product
        # of rotor_frequency * count
        # count = method.spectral_dimensions[0].count
        # if method.spectral_dimensions[0].spectral_width != rotor_frequency * count:
        #     raise ValueError(
        #         "The spectral width along the dimension at index 0 must be equal to "
        #         "the product of the `rotor_frequency` and dimension `count`."
        #     )

        return cls.update(method)

    @staticmethod
    def update(method):
        # setting transition symmetry elements
        for dim in method.spectral_dimensions:
            dim.events[0].transition_query.P = {"channel-1": [[-1]]}
            dim.events[0].transition_query.D = {"channel-1": [[0]]}

        # method affine matrix
        if method.affine_matrix is None:
            method.affine_matrix = [1, -1, 0, 1]

        # method description
        method.description = "Simulate a 2D sideband separation method."

        return method



# -*- coding: utf-8 -*-
"""The Event class."""
from typing import ClassVar
from typing import Dict
from typing import Union

import numpy as np
from pydantic import validator

from ._base import ModuleOperation
from .utils import _get_broadcast_shape
from .utils import _str_to_quantity
from .utils import CONST

__author__ = "Maxwell C. Venetos"
__email__ = "maxvenetos@gmail.com"


class Apodization(ModuleOperation):
    dim_index: Union[int, list, tuple] = 0
    dv_index: Union[int, list, tuple] = None  # if none apply to all
    module_name: ClassVar = __name__

    @property
    def function(self):
        return "apodization"

    def operate(self, data):
        """Apply the operation function.

        Args:
            data: A CSDM object.
        """
        dims = data.dimensions
        ndim = len(dims)

        dim_index = self.dim_index
        dim_index = [dim_index] if isinstance(dim_index, int) else dim_index

        for i in dim_index:
            x = self.get_coordinates(dims[i])  # dims[i].coordinates
            apodization_vactor = _get_broadcast_shape(self.fn(x), i, ndim)

            dv_indexes = self._get_dv_indexes(self.dv_index, n=len(data.y))

            for index in dv_indexes:
                data.y[index].components *= apodization_vactor
        return data


class Gaussian(Apodization):
    r"""Apodize a dependent variable of the CSDM object with a Gaussian function.

    The apodization function follows

    .. math::
        f(x) = e^{-2 \pi^2 \sigma^2  x^2},

    where :math:`x` are the coordinates of the dimension, and :math:`\sigma` is the
    standard deviation. The relationship between the standard deviation, :math:`\sigma`,
    and the full width at half maximum of the reciprocal domain Gaussian function
    follows

    .. math::
        \sigma = \frac{\text{FWHM}}{2\sqrt{2\ln 2}}.

    Args:
        str FWHM: The full width at half maximum, FWHM, of the reciprocal domain
            Gaussian function, given as a string with a value and a unit. The default
            value is 0.
        int dim_index: The index of the CSDM dimension along which the operation is
            applied. The default is the dimension at index 0.
        int dv_index: The index of the CSDM dependent variable where the operation is
            applied. If the value is None, the operation will be applied to every
            dependent variable.

    Example
    -------

    >>> operation4 = sp.apodization.Gaussian(FWHM='143.4 Hz', dim_index=0, dv_index=0)
    """

    FWHM: Union[float, str] = 0
    property_units: Dict = {"FWHM": CONST}

    @validator("FWHM")
    def str_to_quantity(cls, v, values):
        return _str_to_quantity(v, values, "FWHM")

    def fn(self, x):
        x = self.get_coordinates_in_units(x, unit=1.0 / self.property_units["FWHM"])
        sigma = self.FWHM / 2.354820045030949
        return 1.0 if self.FWHM == 0.0 else np.exp(-2.0 * (np.pi * sigma * x) ** 2)


class Exponential(Apodization):
    r"""Apodize a dependent variable of the CSDM object by an exponential function.

    The apodization function follows

    .. math::
        f(x) = e^{-\Gamma \pi |x|},

    where :math:`x` are the coordinates of the dimension, and :math:`\Gamma` is the
    width parameter. The relationship between the width parameter, :math:`\Gamma`, and
    the full width at half maximum for the reciprocal domain Lorentzian function
    follows

    .. math::
        \text{FWHM} = \Gamma.

    Args:
        str FWHM: The full width at half maximum, FWHM, of the reciprocal domain
            Lorentzian function given as a string with a value and a unit. The default
            value is 0.
        int dim_index: The index of the CSDM dimension along which the operation is
            applied. The default is the dimension at index 0.
        int dv_index: The index of the CSDM dependent variable where the operation is
            applied. If the value is None, the operation will be applied to every
            dependent variable.

    Example
    -------

    >>> operation5 = sp.apodization.Exponential(FWHM='143.4 m', dim_index=0, dv_index=0)
    """

    FWHM: Union[float, str] = 0
    property_units: Dict = {"FWHM": CONST}

    @validator("FWHM")
    def str_to_quantity(cls, v, values):
        return _str_to_quantity(v, values, "FWHM")

    def fn(self, x):
        x = self.get_coordinates_in_units(x, unit=1.0 / self.property_units["FWHM"])
        return 1.0 if self.FWHM == 0.0 else np.exp(-self.FWHM * np.pi * np.abs(x))



# -*- coding: utf-8 -*-
import getopt
import sys

from .benchmark import Benchmark


class Main:
    def __init__(self):
        self.benchmark_level = None
        self.n_jobs = 1

    def get_args(self, opts):
        for opt, arg in opts:
            if opt == "-h":  # help
                print("--benchmark=<option=l0,l1,l2>")
                break
            if opt == "--benchmark":  # benchmark
                if int(arg[-1]) > 2:
                    allow = [f"l{i}" for i in range(3)]
                    print(f"Allowed levels are {', '.join(allow)}")
                    sys.exit(2)
                self.benchmark_level = arg
            if opt == "--n_jobs":  # n_jobs
                self.n_jobs = arg

    def benchmark(self):
        if self.benchmark_level is not None:
            getattr(Benchmark, "prep")()
            getattr(Benchmark, self.benchmark_level)(n_jobs=int(self.n_jobs))


if __name__ == "__main__":
    argv = sys.argv[1:]
    try:
        opts, args = getopt.getopt(argv, "h", ["benchmark=", "n_jobs="])
    except getopt.GetoptError:
        print("--benchmark=<option>")
        sys.exit(2)

    start = Main()
    start.get_args(opts)
    start.benchmark()



# -*- coding: utf-8 -*-
"""The Transition class."""
from typing import List

import numpy as np
from pydantic import BaseModel

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class Transition(BaseModel):
    r"""Base Transition class describes a spin transition between two energy states,
    where the energy states are described using the weakly coupled basis.

    .. math::
        |m_{i,0}, m_{i,1}, ... m_{i,N} \rangle \rightarrow
            |m_{f,0}, m_{f,1}, ... m_{f,N} \rangle

    Arguments:
        list initial: The initial Zeeman energy state represented as a list of quantum
            numbers :math:`m_{i,n}`.
        list final: The final Zeeman energy state represented as a list of quantum
            numbers :math:`m_{f,n}`.

    Example:
        >>> from mrsimulator.transition import Transition
        >>> t1 = Transition(initial = [0.5, 0.5], final = [0.5, -0.5])
        >>> t1
        |0.5, -0.5⟩⟨0.5, 0.5|
    """

    initial: List[float] = []
    final: List[float] = []

    def __repr__(self):
        """Representation in bar-ket notation, |final⟩⟨initial|."""
        final = ", ".join([str(i) for i in self.final])
        initial = ", ".join([str(i) for i in self.initial])
        return f"|{final}⟩⟨{initial}|"

    def __str__(self):
        return self.__repr__()

    # @property
    # def Zeeman_allowed(self):
    #     if abs(self.p) == 1:
    #         return True
    #     return False

    @property
    def p(self):
        """Return the total Δm (m_final-m_initial) value of the spin transition.

        Example:
            >>> t1.p
            -1.0
        """
        return self.P.sum()

    @property
    def delta_m(self):
        """An alias for p"""
        return self.p

    @property
    def P(self):
        """Return a list of Δm values of the spin transition for each site in a
        weakly coupled basis.

        Example:
            >>> t1.P
            array([ 0., -1.])
        """
        return np.asarray(self.final) - np.asarray(self.initial)

    # @property
    # def PP(self):
    #     """Return a list of Δm values of the spin transition for each site."""
    #     return np.prod(np.asarray(self.final)) - np.prod(np.asarray(self.initial))

    @property
    def D(self):
        """Return a list of Δm**2 values of the spin transition for each site in a
        weakly coupled basis.

        Example:
            >>> t1.D
            array([0., 0.])
        """
        return np.asarray(self.final) ** 2 - np.asarray(self.initial) ** 2

    # def delta_m(self, i):
    #     """Return the Δm element of the transition corresponding to the ith site."""
    #     return self.final[i] - self.initial[i]

    def tolist(self) -> list:
        """Convert the transition to a list of quantum numbers where the first N
        quantum numbers corresponds to the initial energy state, while the last N
        corresponds to the final energy state, where N is the number of sites.

        Example:
            >>> t1.tolist()
            [0.5, 0.5, 0.5, -0.5]
        """
        lst = self.initial + self.final
        return lst

    def json(self) -> dict:
        """Parse the class object to a JSON compliant python dictionary object.

        Example:
            >>> t1.json()
            {'initial': [0.5, 0.5], 'final': [0.5, -0.5]}
        """
        return self.dict()



# -*- coding: utf-8 -*-
# import csdmpy as cp
# import plotly.graph_objects as go

__author__ = "Deepansh Srivastava"
__email__ = "srivastava.89@osu.edu"


def get_spectral_dimensions(csdm_object, units=False):
    """Extract the count, spectral_width, and reference_offset parameters, associated
    with the spectral dimensions of the method, from the CSDM dimension objects.

    Args:
        csdm_object: A CSDM object holding the measurement dataset.

    Returns:
        A list of dict objects, where each dict containts the count, spectral_width, and
        reference_offset.
    """
    result = []
    for dim in csdm_object.dimensions:
        count = dim.count
        increment = dim.increment.to("Hz").value
        ref = dim.coordinates_offset.to("Hz").value
        sw = count * increment

        even = count % 2 == 0
        complex_co = ref + sw / 2.0
        complex_co -= 0 if even else increment / 2.0
        co = ref if dim.complex_fft else complex_co

        if sw < 0:
            sw = -sw
            co += -increment if even else 0

        dim_i = {}
        dim_i["count"] = dim.count
        dim_i["spectral_width"] = sw if not units else f"{sw} Hz"
        dim_i["reference_offset"] = co if not units else f"{co} Hz"

        if dim.label != "":
            dim_i["label"] = dim.label

        if dim.origin_offset not in ["", None, 0]:
            oo = dim.origin_offset.to("Hz").value
            dim_i["origin_offset"] = oo if not units else f"{oo} Hz"
        result.append(dim_i)

    return result[::-1]


def flatten_dict(obj, previous_key=None):
    """Flatten a nested dictionary with keys as obj1.obj2... and so on"""
    result = {}
    for k, v in obj.items():
        if not isinstance(v, dict):
            key = f"{previous_key}.{k}" if previous_key is not None else k
            result.update({key: v})
        else:
            result.update(**flatten_dict(v, previous_key=k))

    return result


# def plotly_scatter_obj(data: cp.CSDM = None, label: str = None):
#     """Plotly figure dict.
#     Example:

#         # >>> fig_obj = plotly_scatter_obj(experiment, label="experiment")
#         # >>> fig_obj["data"] += plotly_scatter_obj(best_fit)["data"]
#         # >>> fig_obj["data"] += plotly_scatter_obj(residuals, label="res")["data"]
#         # >>> fig = go.Figure(fig_obj)
#         # >>> fig
#     """
#     data_ = [
#         go.Scatter(
#             x=data.x[0].coordinates.value,
#             y=item.components[0],
#             mode="lines",
#             opacity=0.75,
#             name=item.name if label is None else label,
#         )
#         for item in data.y
#     ]
#     layout_ = go.Layout(
#         xaxis=dict(
#             title=data.x[0].axis_label,
#             ticks="outside",
#             showline=True,
#             autorange="reversed",
#             zeroline=False,
#         ),
#         yaxis=dict(
#             ticks="outside",
#             showline=True,
#             zeroline=False,
#             autorange=True,
#         ),
#         margin={"l": 60, "b": 45, "t": 50, "r": 50},
#         width=800,
#         height=400,
#     )

#     return {"data": data_, "layout": layout_}
